#!/usr/bin/python
"""
Empire command-line client tests
"""
import json
import unittest
from empire.client import EmpireException

import httmock

from empire import Empire


class FakeConnection(object):
    def close(self):
        pass


class EmpireTest(unittest.TestCase):
    json_headers = {'content-type': 'application/json'}

    def setUp(self):
        self.empire = Empire(appkey="MOCK_APPKEY", enduser="MOCK_ENDUSER", api_server='api.empire.co')
        self.services = {}

    def mock_response_200(self, request, data):
        headers = {}

        if not isinstance(data, basestring):
            data = json.dumps(data)
            headers = self.json_headers

        response = httmock.response(200, data, headers, None, 5, request)
        setattr(response, 'connection', FakeConnection())
        return response

    def mock_response_500(self, request, message):
        response = httmock.response(500, message, {}, None, 5, request)
        setattr(response, 'connection', FakeConnection())
        return response

    @httmock.urlmatch(netloc='api.empire.co', path='/empire/session/create')
    def session_create_mock(self, url, request):
        return self.mock_response_200(request, {'status': 'OK', 'sessionkey': 'TESTSESSION'})

    @httmock.urlmatch(netloc='api.empire.co', path='/empire/services/salesforce/connect')
    def connect_mock(self, url, request):
        data = json.loads(request.body)
        self.services['salesforce'] = data
        return self.mock_response_200(request, {'status': 'OK'})

    def test_connect(self):
        sf_data = {
            "access_token": "MOCK_ACCESS_TOKEN",
            "client_id": "MOCK_CLIENT",
            "refresh_token": "MOCK_REFRESH_TOKEN",
            "endpoint": "https://na15.salesforce.com"
        }

        with httmock.HTTMock(self.session_create_mock):
            with httmock.HTTMock(self.connect_mock):
                self.empire.connect("salesforce", sf_data)

        self.assertEqual(self.services['salesforce'], sf_data)

    @httmock.urlmatch(netloc='api.empire.co', path='/empire/services')
    def describe_all_mock(self, url, request):
        return self.mock_response_200(request, {'status': 'OK', "name": "salesforce"})

    def test_describe_all(self):
        with httmock.HTTMock(self.session_create_mock):
            with httmock.HTTMock(self.describe_all_mock):
                services_data = self.empire.describe()

        self.assertEqual(services_data, {'status': 'OK', "name": "salesforce"})

    @httmock.urlmatch(netloc='api.empire.co', path='/empire/services/salesforce')
    def describe_one_mock(self, url, request):
        return self.mock_response_200(request, {'status': 'OK', "name": "salesforce", "tables": ["table1"]})

    def test_describe_one(self):
        with httmock.HTTMock(self.session_create_mock):
            with httmock.HTTMock(self.describe_one_mock):
                service_data = self.empire.describe("salesforce")

        self.assertEqual(service_data, {'status': 'OK', "name": "salesforce", "tables": ["table1"]})

    @httmock.urlmatch(netloc='api.empire.co', path='/empire/services/salesforce/table1')
    def describe_table_mock(self, url, request):
        return self.mock_response_200(request, {'status': 'OK', "name": "table1"})

    def test_describe_table(self):
        with httmock.HTTMock(self.session_create_mock):
            with httmock.HTTMock(self.describe_table_mock):
                service_data = self.empire.describe("salesforce", "table1")

        self.assertEqual(service_data, {'status': 'OK', "name": "table1"})

    def test_describe_table_without_service(self):
        self.assertRaises(ValueError, self.empire.describe, None, "table1")

    @httmock.urlmatch(netloc='api.empire.co', path='/empire/services/salesforce/table1')
    def describe_table_broken_mock(self, url, request):
        return self.mock_response_500(request, "Something is broken")

    def test_describe_broken(self):
        with httmock.HTTMock(self.session_create_mock):
            with httmock.HTTMock(self.describe_table_broken_mock):
                self.assertRaises(EmpireException, self.empire.describe, "salesforce", "table1")

    @httmock.urlmatch(netloc='api.empire.co', path='/empire/query')
    def query_mock(self, url, request):
        return self.mock_response_200(request, json.dumps({'col1': 'val1'}) + '\n' + json.dumps({'col2': 'val2'}))

    def test_query(self):
        with httmock.HTTMock(self.session_create_mock):
            with httmock.HTTMock(self.query_mock):
                query_result = list(self.empire.query('SELECT * FROM salesforce.account'))

        self.assertEqual(query_result, [{'col1': 'val1'}, {'col2': 'val2'}])

    @httmock.urlmatch(netloc='api.empire.co', path='/empire/view/viewName')
    def view_create_mock(self, url, request):
        if request.method == 'PUT':
            if json.loads(request.body) == {'query': 'SELECT QUERY'}:
                self.view_created = True
                return self.mock_response_200(request, {'status': 'OK'})

        return self.mock_response_500(request, "Something is broken")

    def test_materialize_view(self):
        self.view_created = False
        self.assertEqual(self.view_created, False)

        with httmock.HTTMock(self.session_create_mock):
            with httmock.HTTMock(self.view_create_mock):
                r = self.empire.materialize_view(name='viewName', sql='SELECT QUERY')

        self.assertEqual(r, {'status': 'OK'})
        self.assertEqual(self.view_created, True)

    @httmock.urlmatch(netloc='api.empire.co', path='/empire/view/viewName')
    def view_delete_mock(self, url, request):
        if request.method == 'DELETE':
            self.view_deleted = True
            return self.mock_response_200(request, {'status': 'OK'})

        return self.mock_response_500(request, "Something is broken")

    def test_drop_view(self):
        self.view_deleted = False
        self.assertEqual(self.view_deleted, False)

        with httmock.HTTMock(self.session_create_mock):
            with httmock.HTTMock(self.view_delete_mock):
                r = self.empire.drop_view(name='viewName')

        self.assertEqual(r, {'status': 'OK'})
        self.assertEqual(self.view_deleted, True)




from distutils.core import setup

setup(
    name='empire-client',
    version='0.3.3',
    author='UPSHOT Data, Inc.',
    author_email='hello@empiredata.co',
    url='http://empiredata.co',
    description='Python client for the Empire API',
    long_description='Empire is an API for accessing enterprise SaaS services such as Salesforce, Zendesk, Google Apps, etc. It provides a uniform, database-like interface to every service that it supports. Empire makes it easy to integrate data from multiple enterprise services into your own enterprise app.',
    download_url='https://github.com/empiredata/empire-python-client',
    license='Apache License, Version 2.0',
    packages=['empire'],
    install_requires=[
        'requests',
        'pyaml',
        'pager',
        'python-dateutil'
    ],
    test_suite='nose.collector',
)



#!/usr/bin/python
"""
Empire command-line client
"""

import sys
import json
import urllib
import time

import requests
import pyaml
import pager
import dateutil.parser

class EmpireException(Exception):
    pass

class Empire(object):
    def __init__(self, appkey=None, enduser=None, api_server='api.empiredata.co', secrets_yaml=None):
        """
        appkey is your Empire application key, and is necessary for using the API.
        enduser is an optional string, identifying the enduser. It is necessary when doing any operation on a view.
        """
#    def __init__(self, appkey=None, api_server='localhost:9000', secrets_yaml=None):
        self.appkey = appkey
        self.enduser = enduser
        self.api_server = api_server
        if self.api_server.startswith("localhost"):
            self.base_url = "http://%s/empire/" % self.api_server
        else:
            self.base_url = "https://%s/empire/" % self.api_server
        self.stream_chunk_size = 128  # This is set low because lines aren't very long.
        if secrets_yaml:
            d = pyaml.yaml.load(open(secrets_yaml))
            self.service_secrets = {name: {o: v['value'] for o,v in d[name]['option'].items()} for name in d}
            # Convert all secrets to strings
            # TODO Remove these workarounds
            for s in self.service_secrets:
                if "scope" in self.service_secrets[s]:
                    del self.service_secrets[s]["scope"]
                for k in self.service_secrets[s]:
                    self.service_secrets[s][k] = str(self.service_secrets[s][k])
        else:
            self.service_secrets = None
        self.sessionkey = None

    def connect(self, service, secrets=None):
        url = self.base_url + 'services/%s/connect' % service
        if not secrets:
            if not self.service_secrets:
                raise ValueError("secrets must be provided on connect command, or secrets_yaml file must be given "
                                 "when constructing this instance")
            secrets = self.service_secrets[service]

            # TODO: Better err if self.service_secrets doesn't exist
        data = json.dumps(secrets)
        return self._do_request('post', url, data=data)

    def describe(self, service=None, table=None):
        url = self.base_url + "services"
        if service and table:
            url += '/%s/%s' % (service, table)
        elif service and not table:
            url += '/%s' % service
        elif not service and table:
            raise ValueError("Service must be specified if table is specified")

        return self._do_request('get', url)

    def print_query(self, sql):
        """
        Paginated printing of an SQL query
        """
        pager.page(self.query(sql, decode=False))

    def query(self, sql, decode=True):
        """
        Issue a SQL query, with an iterator over row results
        """
        url = self.base_url + "query"
        data = json.dumps({"query": sql})

        r = self._do_request('post', url, data=data, stream=True, intact=True)
        try:
            for l in r.iter_lines(chunk_size=self.stream_chunk_size):
                if decode:
                    yield json.loads(l)
                else:
                    yield l
        finally:
            if hasattr(r, 'connection'):
                r.connection.close()

    def insert(self, service, table, row):
        """
        Insert a new row into this service table.
        The row should be a dict of {column: value}
        """
        url = self.base_url + "services/%s/%s" % (service, table)
        data = json.dumps(row)
        return self._do_request('post', url, data=data)

    def materialize_view(self, name, sql):
        """
        Materialize a SQL query as a view. This creates or updates a view.
        """
        if not self.enduser:
            raise EmpireException("Cannot use materialized view within a session initiated without an enduser")
        url = self.base_url + "view/%s" % name
        data = json.dumps({"query": sql})
        return self._do_request('put', url, data=data)

    def drop_view(self, name):
        """
        Delete a materialized view of SQL query
        """
        if not self.enduser:
            raise EmpireException("Cannot use materialized view within a session initiated without an enduser")
        url = self.base_url + "view/%s" % name
        return self._do_request('delete', url)

    def view_ready(self, name):
        """
        Boolean check if a materialized view is ready for querying.
        @note The user is expected to check view_ready before calling query()
        """
        r = self._view_status(name)
        if r["viewStatus"] == "ready":
            return True
        elif r["viewStatus"] == "pending":
            return False
        else:
            raise EmpireException("Unknown view status: %s" % r["viewStatus"])

    def view_materialized_at(self, name):
        """
        Datetime that this view was materialized at. None if the
        materialization is currently pending.
        """
        r = self._view_status(name)
        if "materializedAt" in r:
            return dateutil.parser.parse(r["materializedAt"])
        else:
            return None

    def _view_status(self, name):
        if not self.enduser:
            raise EmpireException("Cannot use materialized view within a session initiated without an enduser")
        url = self.base_url + "view/%s/status" % name
        return self._do_request('get', url)

    def walkthrough(self):
        self.lastservice = None
        self.lasttable = None
        if not self.service_secrets:
            print "Please connect some services in https://login.empiredata.co, and download the new yaml file"
            return

        for service in self.service_secrets:
            self._walkthrough_service(service)

        self._walkthrough_materialized_view(self.lastservice, self.lasttable)

    def _walkthrough_service(self, service):
        try:
            print "empire.connect(%s)" % repr(service)
            self.connect(service)
    
            for table_data in self.describe(service)["service"]["tables"]:
                table = table_data["table"]
                self._walkthrough_table(service, table)
        except Exception, e:
            print "Error in service", service, e

    def _walkthrough_table(self, service, table):
        # These mailchimp tables can only be queried
        # when filtering by a particular list.
        if service == "mailchimp":
            skip_mc_tables = ["list_member", "campaign", "campaign_sent_to", "campaign_opened"]
            if table in skip_mc_tables:
                return

        sql = "SELECT * FROM %s.%s LIMIT 5" % (service, table)
        # We print the repr so that strings are correctly quoted
        print "empire.query(%s)" % repr(str(sql))
        for row in self.query(sql):
            self._walkthrough_print_row(row)
        self.lastservice, self.lasttable = service, table

    def _walkthrough_materialized_view(self, lastservice, lasttable):
        if not self.enduser:
            print "Please specify an enduser parameter when instantiating the client, so that you can try materialized views"
            return

        print 'empire.materialize_view("view_name", "SELECT * FROM %s.%s LIMIT 5")' % (lastservice, lasttable)
        self.materialize_view("view_name", "SELECT * FROM %s.%s LIMIT 5" % (lastservice, lasttable))
        print 'while not empire.view_ready("view_name"):\n    time.sleep(0.01)'
        while not self.view_ready("view_name"):
            time.sleep(0.01)
        print 'empire.query("SELECT * FROM view_name")'
        for row in self.query("SELECT * FROM view_name"):
            self._walkthrough_print_row(row)
        print 'empire.drop_view("view_name")'
        self.drop_view("view_name")

    def _walkthrough_print_row(self, row, max_length=70):
        row = repr(row)
        fragment = row[0:max_length]
        if len(fragment) == max_length:
            fragment = fragment + "..."
        print "    %s" % fragment

    def _do_request(self, *args, **kwargs):
        self._ensure_session()
        headers = {'Authorization': 'Empire sessionkey="%s"' % self.sessionkey}
        kwargs.setdefault('headers', {})
        # Does this overwrite the kwargs, in case the user wanted that dict later?
        kwargs['headers']['Authorization'] = 'Empire sessionkey="%s"' % self.sessionkey
        return self._do_request_help(*args, **kwargs)

    def _ensure_session(self):
        if not self.sessionkey:
            if self.enduser:
                # Manually construct the query parameter, because POST will
                # put it in the body by default
                url = self.base_url + 'session/create?enduser=%s' % urllib.quote(self.enduser)
            else:
                url = self.base_url + 'session/create'
            headers = {'Authorization': 'Empire appkey="%s"' % self.appkey}
            r = self._do_request_help('post', url, headers=headers)
            self.sessionkey = r["sessionkey"]
            #print >> sys.stderr, "Created session"

    def _do_request_help(self, *args, **kwargs):
        kwargs.setdefault('headers', {})

        # For POST requests with data, specify that they are JSON
        if 'data' in kwargs:
            kwargs['headers'].setdefault('content-type', 'application/json')

        kwargs.setdefault('stream', True)

        intact = kwargs.pop('intact', False)

        r = requests.request(*args, **kwargs)
        if intact:
            return r

        try:
            result = r.json()
            if result["status"] != "OK":
                raise EmpireException(result["error"])
            return result
        except ValueError, e:
            raise EmpireException(repr(e) + " " + r.text)
        finally:
            r.connection.close()
